// Netify Agent Device Discovery Processor
// Copyright (C) 2023 eGloo Incorporated
// <http://www.egloo.ca>

#include "nd-config.hpp"
#include "nd-protos.hpp"
#include "nd-util.hpp"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>

#include <chrono>
#include <fstream>

#include "npp-plugin.hpp"

void nppGlobalConfig::Load(const json &j) {
    Reset();

    auto it = j.find("max_confidence");
    if (it != j.end() && it->type() == json::value_t::number_unsigned)
    {
        max_confidence = it->get<unsigned>();
    }

    it = j.find("max_mdns_services");
    if (it != j.end() && it->type() == json::value_t::number_unsigned)
    {
        max_mdns_services = it->get<unsigned>();
    }

    it = j.find("max_http_user_agents");
    if (it != j.end() && it->type() == json::value_t::number_unsigned)
    {
        max_http_user_agents = it->get<unsigned>();
    }

    it = j.find("max_ssdp_user_agents");
    if (it != j.end() && it->type() == json::value_t::number_unsigned)
    {
        max_ssdp_user_agents = it->get<unsigned>();
    }

    it = j.find("netify_api");

    if (it != j.end()) {
        auto it_api = it->find("enable");
        if (it_api != it->end() &&
          it_api->type() == json::value_t::boolean)
        {
            api_enabled = it_api->get<bool>();
        }

        it_api = it->find("url");
        if (it_api != it->end() &&
          it_api->type() == json::value_t::string)
        {
            api_url = it_api->get<string>();
        }

        it_api = it->find("key");
        if (it_api != it->end() &&
          it_api->type() == json::value_t::string)
        {
            api_key = it_api->get<string>();
        }
    }
}

void nppChannelConfig::Load(const string &channel, const json &jconf) {
    auto it = jconf.find("format");
    if (it != jconf.end() && it->type() == json::value_t::string)
    {
        string format = it->get<string>();
        if (format == "json")
            this->format = nppChannelConfig::FORMAT_JSON;
        else if (format == "msgpack")
            this->format = nppChannelConfig::FORMAT_MSGPACK;
        else
            throw ndPluginException("format", strerror(EINVAL));
    }

    it = jconf.find("compressor");
    if (it != jconf.end() && it->type() == json::value_t::string)
    {
        string compressor = it->get<string>();
        if (compressor == "none")
            this->compressor = nppChannelConfig::COMPRESSOR_NONE;
        else if (compressor == "gz")
            this->compressor = nppChannelConfig::COMPRESSOR_GZ;
        else {
            throw ndPluginException("compressor", strerror(EINVAL));
        }
    }
}

void nppDevice::Update(const nppGlobalConfig &config,
  const nppDevice &device) {

    ip_list.insert(device.ip_list.begin(), device.ip_list.end());

    if (! device.dhcp_fingerprint.empty() &&
      device.dhcp_fingerprint != dhcp_fingerprint)
    {
        updates++;
        dhcp_fingerprint = device.dhcp_fingerprint;
    }
    if (! device.dhcp_vendor.empty() && device.dhcp_vendor != dhcp_vendor)
    {
        updates++;
        dhcp_vendor = device.dhcp_vendor;
    }
    if (! device.dhcp_hostname.empty() &&
      device.dhcp_hostname != dhcp_hostname)
    {
        updates++;
        dhcp_hostname = device.dhcp_hostname;
    }

    if (! device.mdns_services.empty()) {
        bool unique = true;
        for (auto &i : mdns_services) {
            if (i != device.mdns_services.front()) continue;
            unique = false;
            break;
        }

        if (unique) {
            updates++;
            mdns_services.push_front(device.mdns_services.front());
            if (mdns_services.size() > config.max_mdns_services)
                mdns_services.pop_back();
        }
    }

    if (! device.http_user_agents.empty()) {
        bool unique = true;
        for (auto &i : http_user_agents) {
            if (i != device.http_user_agents.front())
                continue;
            unique = false;
            break;
        }

        if (unique) {
            updates++;
            http_user_agents.push_front(
              device.http_user_agents.front());
            if (http_user_agents.size() > config.max_http_user_agents)
                http_user_agents.pop_back();
        }
    }

    if (! device.ssdp_user_agents.empty()) {
        bool unique = true;
        for (auto &i : ssdp_user_agents) {
            if (i != device.ssdp_user_agents.front())
                continue;
            unique = false;
            break;
        }

        if (unique) {
            updates++;
            ssdp_user_agents.push_front(
              device.ssdp_user_agents.front());
            if (ssdp_user_agents.size() > config.max_ssdp_user_agents)
                ssdp_user_agents.pop_back();
        }
    }
}

void nppDevice::Encode(json &j, uint8_t flags) {
    json jdev;

    jdev["mac"] = mac;

    if (flags & ET_DHCP) {
        jdev["dhcp"]["fingerprint"] = dhcp_fingerprint;
        jdev["dhcp"]["vendor"] = dhcp_vendor;
        jdev["dhcp"]["hostname"] = dhcp_hostname;
    }

    if (flags & ET_MDNS)
        jdev["mdns_services"] = mdns_services;
    if (flags & ET_HTTP)
        jdev["http_user_agents"] = http_user_agents;
    if (flags & ET_SSDP)
        jdev["ssdp_user_agents"] = ssdp_user_agents;
    if (flags & ET_CONFIDENCE)
        jdev["confidence"] = confidence;
    //    if (flags & ET_IDENTITY)
    //        jdev["?"] = ?;
    if (flags & ET_IP_COUNT)
        jdev["ip_count"] = ip_list.size();

    j[mac] = jdev;
}

void *nppApiDeviceDiscovery::Entry(void) {

    static map<ndGlobalConfig::UUID, string> uuids = {
        { ndGlobalConfig::UUID_AGENT, "X-UUID" },
        { ndGlobalConfig::UUID_SERIAL, "X-UUID-Serial" },
        { ndGlobalConfig::UUID_SITE, "X-UUID-Site" }
    };

    Headers headers;

    for (auto &uuid : uuids) {
        string value("-");

        if (! ndGC.LoadUUID(uuid.first, value)) {
            nd_dprintf("%s: no UUID set for: %s\n",
              tag.c_str(), uuid.second.c_str());
        }
        headers.insert(make_pair(uuid.second, value));
    }

    if (! mime_type.empty()) {
        headers.insert(make_pair("Content-Type", mime_type));
    }

    try {
        Perform(method, config.api_url, headers, payload);
    }
    catch (CURLcode &rc) {
        nd_dprintf("%s: API request error: %s\n",
          tag.c_str(), curl_easy_strerror(rc));
    }

    return nullptr;
}

nppPlugin::nppPlugin(const string &tag, const ndPlugin::Params &params)
  : ndPluginProcessor(tag, params), reload(false),
    update(false), api_thread(nullptr) {
    if (conf_filename.empty())
        throw ndPluginException("conf_filename", strerror(EINVAL));

    nd_dprintf("%s: initialized\n", tag.c_str());
}

nppPlugin::~nppPlugin() {

    Join();

    if (api_thread != nullptr) {
        delete api_thread;
        api_thread = nullptr;
    }

    nd_dprintf("%s: destroyed\n", tag.c_str());
}

void *nppPlugin::Entry(void) {
    nd_printf(
      "%s: %s v%s Copyright (C) 2021-2023 eGloo "
      "Incorporated.\n",
      tag.c_str(), PACKAGE_NAME, PACKAGE_VERSION);

    Reload();

    if (globals.api_enabled) {
        try {
            api_thread = new nppApiDeviceDiscovery(globals);
            api_thread->Create();
        }
        catch (exception &e) {
            nd_dprintf(
              "%s: failed to create API thread: "
              "%s\n",
              e.what());
        }
    }

    while (! ShouldTerminate()) {
        if (reload.load()) {
            Reload();
            reload = false;
        }

        if (update.load()) {
            ProcessUpdate();
            update = false;
        }

        if (api_thread != nullptr && api_thread->HasTerminated())
        {
            ProcessApiResponse();

            delete api_thread;
            api_thread = nullptr;
        }

        unique_lock<mutex> ul(flow_queue_mutex);
        if (! flow_queue_cond.wait_for(ul, chrono::seconds(1),
              [this] { return (! flow_queue.empty()); }))
        {
            continue;
        }

        ProcessFlow(flow_queue.front());

        flow_queue.pop();
    }

    return NULL;
}

void nppPlugin::DispatchEvent(ndPlugin::Event event, void *param) {
    // General plugin event handling.  Event types are defined in:
    // https://gitlab.com/netify.ai/public/netify-agent/-/blob/master/include/nd-plugin.h
    switch (event) {
    case ndPlugin::EVENT_RELOAD: reload = true; break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndFlowMap *flow_map) {
    // Flow map update event.
    switch (event) {
    case ndPluginProcessor::EVENT_FLOW_MAP: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, nd_flow_ptr &flow) {
    // Individual flow events.
    switch (event) {
    case ndPluginProcessor::EVENT_FLOW_NEW: break;
    case ndPluginProcessor::EVENT_FLOW_EXPIRING: break;
    case ndPluginProcessor::EVENT_FLOW_EXPIRE: break;
    case ndPluginProcessor::EVENT_DPI_NEW: break;
    case ndPluginProcessor::EVENT_DPI_UPDATE: break;
    case ndPluginProcessor::EVENT_DPI_COMPLETE:
    {
        if (flow->flags.ip_nat.load()) break;

        lock_guard<mutex> lg(flow_queue_mutex);
        flow_queue.push(flow);
    }
        flow_queue_cond.notify_one();
        break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndInterfaces *interfaces) {
    // All capture interfaces and their stats are made
    // available on each update interval.
    switch (event) {
    case ndPluginProcessor::EVENT_INTERFACES: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(ndPluginProcessor::Event event,
  const string &iface, ndPacketStats *stats) {
    // Capture interfaces packet stats events are dispatched
    // on each update interval.
    switch (event) {
    case ndPluginProcessor::EVENT_PKT_CAPTURE_STATS: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndPacketStats *stats) {
    // Global (summary) capture packet stats are made
    // available on each update interval.
    switch (event) {
    case ndPluginProcessor::EVENT_PKT_GLOBAL_STATS: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(
  ndPluginProcessor::Event event, ndInstanceStatus *status) {
    // Indicates that an update interval has expired, and an
    // update period has begun.  Useful to synchronize state,
    // or clear/reset collected data.
    switch (event) {
    case ndPluginProcessor::EVENT_UPDATE_INIT: break;
    default: break;
    }
}

void nppPlugin::DispatchProcessorEvent(ndPluginProcessor::Event event) {
    // Indicates that an update period is ending.  Useful to
    // synchronize state, clear/reset collected data, or
    // dispatch a payload to a sink plugin.
    switch (event) {
    case ndPluginProcessor::EVENT_UPDATE_COMPLETE:
        update = true;
        break;
    default: break;
    }
}

void nppPlugin::Reload(void) {
    nd_dprintf("%s: Loading configuration: %s\n",
      tag.c_str(), conf_filename.c_str());

    json j;
    ifstream ifs(conf_filename);
    if (! ifs.is_open()) {
        nd_printf(
          "%s: Error loading configuration: %s: %s\n",
          tag.c_str(), conf_filename.c_str(), strerror(ENOENT));
        Unlock();
        throw ndPluginException("conf_filename", strerror(ENOENT));
    }

    try {
        ifs >> j;
    }
    catch (exception &e) {
        nd_printf(
          "%s: Error loading configuration: %s: JSON parse "
          "error\n",
          tag.c_str(), conf_filename.c_str());
        nd_dprintf("%s: %s: %s\n", tag.c_str(),
          conf_filename.c_str(), e.what());
        Unlock();
        throw ndPluginException("conf_filename", strerror(EINVAL));
    }

    globals.Load(j);

    defaults.Load("defaults", j);

    sinks.clear();

    try {
        auto jsinks = j.find("sinks");

        if (jsinks != j.end()) {
            for (auto &jsink : jsinks->get<json::object_t>())
            {
                for (auto &jchannel :
                  jsink.second.get<json::object_t>())
                {
                    auto it = jchannel.second.find(
                      "enable");
                    if (it != jchannel.second.end() &&
                      it->type() == json::value_t::boolean &&
                      it->get<bool>() != true)
                        continue;

                    nppChannelConfig config;

                    config.Load(jchannel.first,
                      jchannel.second, defaults);

                    auto sink = sinks.find(jsink.first);
                    if (sink == sinks.end()) {
                        map<string, nppChannelConfig> entry;

                        entry.insert(
                          make_pair(jchannel.first, config));
                        sinks.insert(make_pair(jsink.first, entry));
                    }
                    else {
                        sink->second.insert(
                          make_pair(jchannel.first, config));
                    }
                }
            }
        }
    }
    catch (exception &e) {
        Unlock();
        throw e;
    }

    Unlock();
}

void nppPlugin::ProcessFlow(nd_flow_ptr &flow) {
    bool has_metadata = false;
    nppDevice device;
    string origin_ip;

    if (flow->origin == ndFlow::ORIGIN_LOWER) {
        device.mac = flow->lower_mac.GetString();
        origin_ip = flow->lower_addr.GetString();
    }
    else if (flow->origin == ndFlow::ORIGIN_UPPER) {
        device.mac = flow->upper_mac.GetString();
        origin_ip = flow->upper_addr.GetString();
    }
    else {
        nd_dprintf("%s: unknown origin.\n", tag.c_str());
        return;
    }

    if (identified.find(device.mac) != identified.end()) {
        // Device identified, ignore...
        return;
    }

    device.ip_list.insert(origin_ip);

    if (flow->detected_protocol == ND_PROTO_DHCP) {
        if (flow->HasDhcpClassIdent()) {
            has_metadata = true;
            device.dhcp_vendor = flow->dhcp.class_ident;
        }
        if (flow->HasDhcpFingerprint()) {
            has_metadata = true;
            device.dhcp_fingerprint = flow->dhcp.fingerprint;
        }
        if (! flow->host_server_name.empty()) {
            has_metadata = true;
            device.dhcp_hostname = flow->host_server_name;
        }
    }
    if (flow->HasHttpUserAgent()) {
        has_metadata = true;
        device.http_user_agents.emplace_front(flow->http.user_agent);
    }
    if (flow->HasSSDPUserAgent()) {
        has_metadata = true;
        device.ssdp_user_agents.emplace_front(flow->http.user_agent);
    }
    if (flow->HasMDNSDomainName()) {
        has_metadata = true;
        device.mdns_services.emplace_front(flow->mdns.domain_name);
    }

    if (has_metadata) {
        if (ndGC_DEBUG && ndGC.verbosity > 1)
            flow->Print(ndFlow::PRINTF_ALL, tag);

        Devices::iterator i = devices.find(device.mac);
        if (i != devices.end()) {
            // Update device entry...
            i->second.Update(globals, device);
            return;
        }

        // New device entry...
        devices.emplace(device.mac, device);
    }
}

void nppPlugin::PrintDeviceDebug(const nppDevice &device) const {
    ndDebugLogStream dls;

    size_t indent = device.mac.size();
    size_t const max_len = 68;

    dls
      << setw(indent) << device.mac << setw(0) << ":"
      << " Confidence: " << device.confidence << " "
      << ((device.confidence >= globals.max_confidence &&
            ! device.label.empty()) ?
             device.label :
             "")
      << endl
      << setw(indent + 1) << ":" << setw(0)
      << " IP count: " << device.ip_list.size() << endl
      << setw(indent + 1) << ":" << setw(0)
      << " DHCP/FP: " << device.dhcp_fingerprint
      << " DHCP/V: " << device.dhcp_vendor
      << " DHCP/H: " << device.dhcp_hostname << endl
      << setw(indent + 1) << ":" << setw(0)
      << " MDNS services: " << device.mdns_services.size() << endl;

    for (auto &s : device.mdns_services) {
        dls << setw(indent + 1) << ":" << setw(0) << "  - "
            << s.substr(0, max_len)
            << ((s.size() > max_len) ? "..." : "") << endl;
    }

    dls << setw(indent + 1) << ":" << setw(0) << " HTTP user agents: "
        << device.http_user_agents.size() << endl;

    for (auto &s : device.http_user_agents) {
        dls << setw(indent + 1) << ":" << setw(0) << "  - "
            << s.substr(0, max_len)
            << ((s.size() > max_len) ? "..." : "") << endl;
    }

    dls << setw(indent + 1) << ":" << setw(0) << " SSDP user agents: "
        << device.ssdp_user_agents.size() << endl;

    for (auto &s : device.ssdp_user_agents) {
        dls << setw(indent + 1) << ":" << setw(0) << "  - "
            << s.substr(0, max_len)
            << ((s.size() > max_len) ? "..." : "") << endl;
    }

    dls << endl;
}

void nppPlugin::ProcessUpdate(void) {
    json jdev;

    for (auto i = devices.begin(); i != devices.end();) {
        if (i->second.confidence >= globals.max_confidence) {
            if (ndGC_DEBUG /* && ndGC_VERBOSE*/)
                PrintDeviceDebug(i->second);

            identified.insert(i->first);

            i = devices.erase(i);
            continue;
        }

        if (i->second.updates) {
            if (ndGC_DEBUG /* && ndGC_VERBOSE*/)
                PrintDeviceDebug(i->second);

            i->second.Encode(jdev,
              nppDevice::ET_METADATA | nppDevice::ET_IP_COUNT);
            i->second.updates = 0;
        }

        i++;
    }

    if (! jdev.empty()) {
        if (! sinks.empty()) DispatchPayload(jdev);

        if (globals.api_enabled && api_thread == nullptr) {
            try {
                api_thread = new nppApiDeviceDiscovery(globals, jdev);
                api_thread->Create();
            }
            catch (exception &e) {
                nd_dprintf(
                  "%s: failed to create API thread: "
                  "%s\n",
                  e.what());
            }
        }
    }
}

void nppPlugin::DispatchPayload(json &j) {
    for (auto &sink : sinks) {
        for (auto &channel : sink.second) {
            uint8_t flags = ndPlugin::DF_NONE;

            switch (channel.second.format) {
            case nppChannelConfig::FORMAT_JSON:
                flags |= ndPlugin::DF_FORMAT_JSON;
                break;
            case nppChannelConfig::FORMAT_MSGPACK:
                flags |= ndPlugin::DF_FORMAT_MSGPACK;
                break;
            default: break;
            }

            switch (channel.second.compressor) {
            case nppChannelConfig::COMPRESSOR_GZ:
                flags |= ndPlugin::DF_GZ_DEFLATE;
                break;
            default: break;
            }

            DispatchSinkPayload(sink.first,
              { channel.first }, j, flags);
        }
    }
}

void nppPlugin::ProcessApiResponse(void) {
    if (api_thread->http_rc != 200) {
        nd_dprintf("%s: unexpected HTTP response: %lu\n",
          tag.c_str(), api_thread->http_rc);
    }

    if (api_thread->content_type != "application/json") {
        nd_dprintf("%s: unexpected content-type: %s\n",
          tag.c_str(),
          (api_thread->content_type.empty()) ?
            "none" :
            api_thread->content_type.c_str());
        return;
    }

    if (! api_thread->content.size()) {
        nd_dprintf("%s: no content returned.\n", tag.c_str());
        return;
    }

    json result;
    try {
        result = json::parse(api_thread->content);
    }
    catch (exception &e) {
        nd_dprintf("%s: malformed JSON: %s\n", tag.c_str(),
          e.what());
        return;
    }

    auto it = result.find("status_code");
    if (it == result.end() || it->type() != json::value_t::number_unsigned)
    {
        nd_dprintf("%s: %s not found or wrong type.\n",
          tag.c_str(), "status_code");
        return;
    }

    unsigned status_code = it->get<unsigned>();

    it = result.find("status_message");
    if (it == result.end() || it->type() != json::value_t::string)
    {
        nd_dprintf("%s: %s not found or wrong type.\n",
          tag.c_str(), "status_message");
        return;
    }

    string status_message = it->get<string>();

    if (status_code != 0) {
        nd_printf("%s: API error: %s [%d]\n", tag.c_str(),
          status_message.c_str(), status_code);
        return;
    }

    it = result.find("data");
    if (it == result.end() || it->type() != json::value_t::array)
    {
        nd_dprintf("%s: %s not found or wrong type.\n",
          tag.c_str(), "data");
        return;
    }

    for (auto &j : it->get<json::array_t>()) {
        auto jmac = j.find("mac");
        if (jmac == j.end() || jmac->type() != json::value_t::string)
        {
            nd_dprintf("%s: %s not found or wrong type.\n",
              tag.c_str(), "mac");
            return;
        }

        string mac = jmac->get<string>();

        auto jdisc = j.find("discovery");
        if (jdisc == j.end() || jdisc->type() != json::value_t::object)
        {
            nd_dprintf("%s: %s not found or wrong type.\n",
              tag.c_str(), "discovery");
            return;
        }

        auto jlabel = jdisc->find("label");
        if (jlabel == jdisc->end() ||
          jlabel->type() != json::value_t::string)
        {
            nd_dprintf("%s: %s not found or wrong type.\n",
              tag.c_str(), "label");
            return;
        }

        string label = jlabel->get<string>();

        auto jscore = jdisc->find("score");
        if (jscore == jdisc->end() ||
          jscore->type() != json::value_t::number_unsigned)
        {
            nd_dprintf("%s: %s not found or wrong type.\n",
              tag.c_str(), "score");
            return;
        }

        unsigned score = jscore->get<unsigned>();

        nd_dprintf("%s: %s: %s: score: %u\n", tag.c_str(),
          mac.c_str(),
          (label.empty()) ? "<NO LABEL>" : label.c_str(), score);

        auto device = devices.find(mac);
        if (device != devices.end()) {
            device->second.label = label;
            device->second.confidence = score;
        }
    }
}

ndPluginInit(nppPlugin);
